import click
from curator.cli_singletons.object_class import cli_action
from curator.cli_singletons.utils import get_width, json_to_dict, validate_filter_json

@click.command(context_settings=get_width())
@click.option('--name', type=str, help='Alias name', required=True)
@click.option('--max_age', type=str, help='max_age condition value (see documentation)')
@click.option('--max_docs', type=str, help='max_docs condition value (see documentation)')
@click.option('--max_size', type=str, help='max_size condition value (see documentation)')
@click.option('--extra_settings', type=str, help='JSON version of extra_settings (see documentation)', callback=json_to_dict)
@click.option('--new_index', type=str, help='Optional new index name (see documentation)')
@click.option('--wait_for_active_shards', type=int, default=1, show_default=True, help='Wait for number of shards to be active before returning')
@click.option('--allow_ilm_indices/--no-allow_ilm_indices', help='Allow Curator to operate on Index Lifecycle Management monitored indices.', default=False, show_default=True)
@click.pass_context
def rollover(ctx, name, max_age, max_docs, max_size, extra_settings, new_index, wait_for_active_shards, allow_ilm_indices):
    """
    Rollover Index associated with Alias
    """
    conditions = {}
    for cond in ['max_age', 'max_docs', 'max_size']:
        if eval(cond) is not None:
            conditions[cond] = eval(cond)
    manual_options = { 
        'name': name,
        'conditions': conditions,
        'allow_ilm_indices': allow_ilm_indices,
    }
    # ctx.info_name is the name of the function or name specified in @click.command decorator
    action = cli_action(
        ctx.info_name, ctx.obj['config']['client'], manual_options, [], True,
        extra_settings=extra_settings,
        new_index=new_index,
        wait_for_active_shards=wait_for_active_shards
    )
    action.do_singleton_action(dry_run=ctx.obj['dry_run'])


import click
import json
import logging
import sys
from curator.actions import Alias, Allocation, Close, ClusterRouting, CreateIndex, DeleteIndices, DeleteSnapshots, ForceMerge, Open, Replicas, Restore, Shrink, Snapshot
from curator.exceptions import NoIndices, NoSnapshots
from curator.validators import SchemaCheck, config_file, options
from curator.utils import ensure_list, filters, prune_nones, validate_filters
from voluptuous import Schema



CLASS_MAP = {
    'alias' :  Alias,
    'allocation' : Allocation,
    'close' : Close,
    'cluster_routing' : ClusterRouting,
    'create_index' : CreateIndex,
    'delete_indices' : DeleteIndices,
    'delete_snapshots' : DeleteSnapshots,
    'forcemerge' : ForceMerge,
    'open' : Open,
    'replicas' : Replicas,
    'restore' : Restore,
    'snapshot' : Snapshot,
}

EXCLUDED_OPTIONS = [
    'ignore_empty_list', 'timeout_override',
    'continue_if_exception', 'disable_action'
]

def get_width():
    return dict(max_content_width=click.get_terminal_size()[0])

def json_to_dict(ctx, param, value):
    if value is None:
        return {}
    try:
        return json.loads(value)
    except ValueError:
        raise click.BadParameter('Invalid JSON for {0}: {1}'.format(param, value))

def validate_filter_json(ctx, param, value):
    # The "None" edge case should only be for optional filters, like alias add/remove
    if value is None:
        return value
    try:
        filter_list = ensure_list(json.loads(value))
        return filter_list
    except ValueError:
        raise click.BadParameter('Filter list is invalid JSON: {0}'.format(value))

def false_to_none(ctx, param, value):
    try:
        if value:
            return True
        else:
            return None
    except ValueError:
        raise click.BadParameter('Invalid value: {0}'.format(value))

def filter_schema_check(action, filter_dict):
    valid_filters = SchemaCheck(
        filter_dict,
        Schema(filters.Filters(action, location='singleton')),
        'filters',
        '{0} singleton action "filters"'.format(action)
    ).result()
    return validate_filters(action, valid_filters)

def actionator(action, action_obj, dry_run=True):
    logger = logging.getLogger(__name__)
    logger.debug('Doing the singleton "{0}" action here.'.format(action))
    try:
        if dry_run:
            action_obj.do_dry_run()
        else:
            action_obj.do_action()
    except Exception as e:
        if isinstance(e, NoIndices) or isinstance(e, NoSnapshots):
            logger.error(
                'Unable to complete action "{0}".  No actionable items '
                'in list: {1}'.format(action, type(e))
            )
        else:
            logger.error(
                'Failed to complete action: {0}.  {1}: '
                '{2}'.format(action, type(e), e)
            )
        sys.exit(1)
    logger.info('Singleton "{0}" action completed.'.format(action))

def do_filters(list_object, filters, ignore=False):
    logger = logging.getLogger(__name__)
    logger.debug('Running filters and testing for empty list object')
    try:
        list_object.iterate_filters(filters)
        list_object.empty_list_check()
    except (NoIndices, NoSnapshots) as e:
        if isinstance(e, NoIndices):
            otype = 'index'
        else:
            otype = 'snapshot'
        if ignore:
            logger.info(
                'Singleton action not performed: empty {0} list'.format(otype)
            )
            sys.exit(0)
        else:
            logger.error(
                'Singleton action failed due to empty {0} list'.format(otype)
            )
            sys.exit(1)


def prune_excluded(option_dict):
    for k in list(option_dict.keys()):
        if k in EXCLUDED_OPTIONS:
            del option_dict[k]
    return option_dict

def option_schema_check(action, option_dict):
    clean_options = SchemaCheck(
        prune_nones(option_dict),
        options.get_schema(action),
        'options',
        '{0} singleton action "options"'.format(action)
    ).result()
    return prune_excluded(clean_options)

def config_override(ctx, config_dict):
    if config_dict == None:
        config_dict = {}
    for k in ['client', 'logging']:
        if not k in config_dict:
            config_dict[k] = {}
    for k in list(ctx.params.keys()):
        if k in ['dry_run', 'config']:
            pass
        elif k == 'host':
            if 'host' in ctx.params and ctx.params['host'] is not None:
                config_dict['client']['hosts'] = ctx.params[k]
        elif k in ['loglevel', 'logfile', 'logformat']:
            if k in ctx.params and ctx.params[k] is not None:
                config_dict['logging'][k] = ctx.params[k]
        else:
            if k in ctx.params and ctx.params[k] is not None:
                config_dict['client'][k] = ctx.params[k]
    # After override, prune the nones
    for k in ['client', 'logging']:
        config_dict[k] = prune_nones(config_dict[k])
    return SchemaCheck(config_dict, config_file.client(),
        'Client Configuration', 'full configuration dictionary').result()


import logging
import sys
from curator import IndexList, SnapshotList
from curator.actions import (
    Alias, Allocation, Close, ClusterRouting, CreateIndex, DeleteIndices, DeleteSnapshots, ForceMerge,
    IndexSettings, Open, Reindex, Replicas, Restore, Rollover, Shrink, Snapshot
)
from curator.defaults.settings import snapshot_actions
from curator.exceptions import ConfigurationError, NoIndices, NoSnapshots
from curator.validators import SchemaCheck, filters, options
from curator.utils import get_client, prune_nones, validate_filters
from voluptuous import Schema


CLASS_MAP = {
    'alias' :  Alias,
    'allocation' : Allocation,
    'close' : Close,
    'cluster_routing' : ClusterRouting,
    'create_index' : CreateIndex,
    'delete_indices' : DeleteIndices,
    'delete_snapshots' : DeleteSnapshots,
    'forcemerge' : ForceMerge,
    'index_settings' : IndexSettings,
    'open' : Open,
    'reindex' : Reindex,
    'replicas' : Replicas,
    'restore' : Restore,
    'rollover': Rollover,
    'shrink': Shrink,
    'snapshot' : Snapshot,
}

EXCLUDED_OPTIONS = [
    'ignore_empty_list', 'timeout_override',
    'continue_if_exception', 'disable_action'
]

class cli_action():
    """
    Unified class for all CLI singleton actions
    """
    def __init__(self, action, client_args, option_dict, filter_list, ignore_empty_list, **kwargs):
        self.logger = logging.getLogger('curator.cli_singletons.cli_action.' + action)
        self.action = action
        self.repository = kwargs['repository'] if 'repository' in kwargs else None
        if action[:5] != 'show_': # Ignore CLASS_MAP for show_indices/show_snapshots
            try:
                self.action_class = CLASS_MAP[action]
            except KeyError:
                self.logger.critical('Action must be one of {0}'.format(list(CLASS_MAP.keys())))
            self.check_options(option_dict)
        else:
            self.options = option_dict
        # Extract allow_ilm_indices so it can be handled separately.
        if 'allow_ilm_indices' in self.options:
            self.allow_ilm = self.options.pop('allow_ilm_indices')
        else:
            self.allow_ilm = False
        if action == 'alias':
            self.alias = {
                'name': option_dict['name'],
                'extra_settings': option_dict['extra_settings'],
                'wini': kwargs['warn_if_no_indices'] if 'warn_if_no_indices' in kwargs else False
            }
            for k in ['add', 'remove']:
                if k in kwargs:
                    self.alias[k] = {}
                    self.check_filters(kwargs[k], loc='alias singleton', key=k)
                    self.alias[k]['filters'] = self.filters
                    if self.allow_ilm:
                        self.alias[k]['filters'].append({'filtertype':'ilm'})
        elif action in [ 'cluster_routing', 'create_index', 'rollover']:
            self.action_kwargs = {}
            # No filters for these actions
            if action == 'rollover':
                # Ugh.  Look how I screwed up here with args instead of kwargs,
                # like EVERY OTHER ACTION seems to have... grr
                # todo: fix this in Curator 6, as it's an API-level change.
                self.action_args = (option_dict['name'], option_dict['conditions'])
                for k in ['new_index', 'extra_settings', 'wait_for_active_shards']:
                    self.action_kwargs[k] = kwargs[k] if k in kwargs else None
        else:
            self.check_filters(filter_list)
        self.client = get_client(**client_args)
        self.ignore = ignore_empty_list

    def prune_excluded(self, option_dict):
        for k in list(option_dict.keys()):
            if k in EXCLUDED_OPTIONS:
                del option_dict[k]
        return option_dict

    def check_options(self, option_dict):
        try:
            self.logger.debug('Validating provided options: {0}'.format(option_dict))
            # Kludgy work-around to needing 'repository' in options for these actions
            # but only to pass the schema check.  It's removed again below.
            if self.action in ['delete_snapshots', 'restore']:
                option_dict['repository'] = self.repository
            _ = SchemaCheck(
                prune_nones(option_dict),
                options.get_schema(self.action),
                'options',
                '{0} singleton action "options"'.format(self.action)
            ).result()
            self.options = self.prune_excluded(_)
            # Remove this after the schema check, as the action class won't need it as an arg
            if self.action in ['delete_snapshots', 'restore']:
                del self.options['repository']
        except ConfigurationError as e:
            self.logger.critical('Unable to parse options: {0}'.format(e))
            sys.exit(1)

    def check_filters(self, filter_dict, loc='singleton', key='filters'):
        try:
            self.logger.debug('Validating provided filters: {0}'.format(filter_dict))
            _ = SchemaCheck(
                filter_dict,
                Schema(filters.Filters(self.action, location=loc)),
                key,
                '{0} singleton action "{1}"'.format(self.action, key)
            ).result()
            self.filters = validate_filters(self.action, _)
        except ConfigurationError as e:
            self.logger.critical('Unable to parse filters: {0}'.format(e))
            sys.exit(1)

    def do_filters(self):
        self.logger.debug('Running filters and testing for empty list object')
        if self.allow_ilm:
            self.filters.append({'filtertype':'ilm','exclude':True})
        try:
            self.list_object.iterate_filters({'filters':self.filters})
            self.list_object.empty_list_check()
        except (NoIndices, NoSnapshots) as e:
            otype = 'index' if isinstance(e, NoIndices) else 'snapshot'
            if self.ignore:
                self.logger.info('Singleton action not performed: empty {0} list'.format(otype))
                sys.exit(0)
            else:
                self.logger.error('Singleton action failed due to empty {0} list'.format(otype))
                sys.exit(1)
    
    def get_list_object(self):
        if self.action in snapshot_actions() or self.action == 'show_snapshots':
            self.list_object = SnapshotList(self.client, repository=self.repository)
        else:
            self.list_object = IndexList(self.client)

    def get_alias_obj(self):
        action_obj = Alias(name=self.alias['name'], extra_settings=self.alias['extra_settings'])
        for k in ['remove', 'add']:
            if k in self.alias:
                self.logger.debug(
                    '{0}ing matching indices {1} alias "{2}"'.format(
                        'Add' if k == 'add' else 'Remov', # 0 = "Add" or "Remov"
                        'to' if k == 'add' else 'from', # 1 = "to" or "from"
                        self.alias['name'] # 2 = the alias name
                    )
                )
                self.alias[k]['ilo'] = IndexList(self.client)
                self.alias[k]['ilo'].iterate_filters({'filters':self.alias[k]['filters']})
                f = getattr(action_obj, k)
                f(self.alias[k]['ilo'], warn_if_no_indices=self.alias['wini'])
        return action_obj

    def do_singleton_action(self, dry_run=False):
        self.logger.debug('Doing the singleton "{0}" action here.'.format(self.action))
        try:
            if self.action == 'alias':
                action_obj = self.get_alias_obj()
            elif self.action in [ 'cluster_routing', 'create_index', 'rollover']:
                action_obj = self.action_class(self.client, *self.action_args, **self.action_kwargs)
            else:
                self.get_list_object()
                self.do_filters()
                self.logger.debug('OPTIONS = {0}'.format(self.options))
                action_obj = self.action_class(self.list_object, **self.options)
            try:
                if dry_run:
                    action_obj.do_dry_run()
                else:
                    action_obj.do_action()
            except Exception as e:
                raise e # pass it on?
        except Exception as e:
            self.logger.critical('Failed to complete action: {0}.  {1}: {2}'.format(self.action, type(e), e))
            sys.exit(1)
        self.logger.info('"{0}" action completed.'.format(self.action))


import click
from curator.cli_singletons.object_class import cli_action
from curator.cli_singletons.utils import get_width, validate_filter_json

@click.command(context_settings=get_width())
@click.option('--key', type=str, required=True, help='Node identification tag')
@click.option('--value', type=str, default=None, help='Value associated with --key')
@click.option('--allocation_type', type=click.Choice(['require', 'include', 'exclude']))
@click.option('--wait_for_completion/--no-wait_for_completion', default=False, help='Wait for the allocation to complete', show_default=True)
@click.option('--max_wait', default=-1, type=int, help='Maximum number of seconds to wait_for_completion', show_default=True)
@click.option('--wait_interval', default=9, type=int, help='Seconds to wait between completion checks.', show_default=True)
@click.option('--ignore_empty_list', is_flag=True, help='Do not raise exception if there are no actionable indices')
@click.option('--allow_ilm_indices/--no-allow_ilm_indices', help='Allow Curator to operate on Index Lifecycle Management monitored indices.', default=False, show_default=True)
@click.option('--filter_list', callback=validate_filter_json, help='JSON array of filters selecting indices to act on.', required=True)
@click.pass_context
def allocation(ctx, key, value, allocation_type, wait_for_completion, max_wait, wait_interval, ignore_empty_list, allow_ilm_indices, filter_list):
    """
    Shard Routing Allocation
    """
    manual_options = {
        'key': key,
        'value': value,
        'allocation_type': allocation_type,
        'wait_for_completion': wait_for_completion,
        'max_wait': max_wait,
        'wait_interval': wait_interval,
        'allow_ilm_indices': allow_ilm_indices,
    }
    # ctx.info_name is the name of the function or name specified in @click.command decorator
    action = cli_action(ctx.info_name, ctx.obj['config']['client'], manual_options, filter_list, ignore_empty_list)
    action.do_singleton_action(dry_run=ctx.obj['dry_run'])


import elasticsearch
import curator
import os
import json
import string, random, tempfile
import click
from click import testing as clicktest
from mock import patch, Mock
from datetime import timedelta, datetime, date

from . import CuratorTestCase
from . import testvars as testvars

import logging
logger = logging.getLogger(__name__)

host, port = os.environ.get('TEST_ES_SERVER', 'localhost:9200').split(':')
port = int(port) if port else 9200

class TestParseDateMath(CuratorTestCase):
    def test_function_positive(self):
        test_string = u'<.prefix-{2001-01-01-13||+1h/h{YYYY-MM-dd-HH|-07:00}}-suffix>'
        expected = u'.prefix-2001-01-01-14-suffix'
        self.assertEqual(expected, curator.parse_datemath(self.client, test_string))
    def test_assorted_datemaths(self):
        for test_string, expected in [
            (u'<prefix-{now}-suffix>', u'prefix-{0}-suffix'.format(datetime.utcnow().strftime('%Y.%m.%d'))),
            (u'<prefix-{now-1d/d}>', u'prefix-{0}'.format((datetime.utcnow()-timedelta(days=1)).strftime('%Y.%m.%d'))),
            (u'<{now+1d/d}>', u'{0}'.format((datetime.utcnow()+timedelta(days=1)).strftime('%Y.%m.%d'))),
            (u'<{now+1d/d}>', u'{0}'.format((datetime.utcnow()+timedelta(days=1)).strftime('%Y.%m.%d'))),
            (u'<{now+10d/d{YYYY-MM}}>', u'{0}'.format((datetime.utcnow()+timedelta(days=10)).strftime('%Y-%m'))),
            (u'<{now+10d/h{YYYY-MM-dd-HH|-07:00}}>', u'{0}'.format((datetime.utcnow()+timedelta(days=10)-timedelta(hours=7)).strftime('%Y-%m-%d-%H'))),
          ]:
            self.assertEqual(expected, curator.parse_datemath(self.client, test_string))